from optel.datalake import cluster_events
# import pytest

project = "optel-data-lab"
zone = "us-east1-d"
cluster = "cluster-datalab-01"


# @pytest.mark.skip(reason="Needs GC credentials to test this.")
def test_get_instances_names(gc_instances):
    instances = cluster_events.get_instances_names(project, zone, cluster)
    print(instances[1])
    assert "runner" not in instances
    assert instances[1] != ""


# @pytest.mark.skip(reason="Needs GC credentials to test this.")
def test_boot_cluster(instances_names):
    cluster_events.boot_cluster(instances_names, project, zone)


def test_shutdown_cluster(instances_names):
    cluster_events.shutdown_cluster(instances_names, project, zone, cluster)



import pyspark.sql.functions as sf
from optel.datalake import transform
import logging


def real_dups(df):
    """
    Check and remove duplicated rows.

    Args:
        df (pyspark.sql.DataFrame): A spark dataframe
    """
    nb_rows = df.count()
    nb_distinct = df.distinct().count()
    nb_dups = nb_rows - nb_distinct
    if nb_rows != nb_distinct:
        logging.info("%s real duplicates found, removing them" % nb_dups)
        dedup_df = df.drop_duplicates()
    else:
        logging.info("No real duplicates found")
        dedup_df = df
    return dedup_df


def empty_columns(df):
    """
    Remove columns with 100% missing values.

    Args:
        df (pyspark.sql.DataFrame): A spark dataframe

    Returns:
        pyspark.sql.DataFrame: A data frame free of any empty columns.
    """
    missing_obs = df.agg(*[(1 - (sf.count(c) / sf.count('*'))).
                           alias(c) for c in df.columns])
    values = missing_obs.first()
    dict_values = values.asDict()
    logging.info(
        "Percentage of empties in each columns %s", dict_values)
    empty_cols = transform.find_key(dict_values, 1.0)
    logging.info("Empty columns found %s", empty_cols)
    for column in empty_cols:
        df = df.drop(column)
    return df


def convert_decimal_to_float(df):
    """
    Convert columns with decimal types to floating type.

    Args:
        df (pyspark.sql.DataFrame): A spark dataframe.

    Returns:
        pyspark.sql.DataFrame: A dataframe free of any decimal type column.

    """
    names = [c[0] for c in df.dtypes if 'decimal' in c[1]]
    return_df = df
    for name in names:
        return_df = return_df.withColumn(name, return_df[name].cast("double"))
    return return_df


def convert_date_to_string(df):
    """
    Convert columns with timestamp and date types to string type with
    the following formatting: yyyy/MM/dd

    Args:
        df (pyspark.sql.DataFrame): A spark dataframe.

    Returns:
        pyspark.sql.DataFrame: A dataframe with dates converted to string

    """

    names = [c[0] for c in df.dtypes if "timestamp" in c[1] or "date" in c[1]]
    print(df.dtypes)
    return_df = df
    for name in names:
        return_df = return_df. \
            withColumn(name, sf.date_format(return_df[name], 'yyyy/MM/dd'))
    return return_df



import pytest
import decimal
import datetime
from optel.datalake.config.sparksession import default_spark
from pyspark.sql.functions import lit
from pyspark.sql.functions import StringType
from pyspark.sql.functions import desc
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType
from pyspark.sql.types import DecimalType
from pyspark.sql.types import TimestampType
from pyspark.sql.types import DateType


@pytest.fixture(scope='session')
def spark_session():
    session = default_spark()
    yield session


@pytest.fixture(scope="session")
def me_df_timestamp(spark_session):
    """Data frame with a unix timestamp. An emtpy column is present
    """

    columns = [
        'date', 'string', 'number',
        'CREATEDTIME', 'COMPLETEDTIME', 'STATUSNAME', 'date2']
    vals = [
        (1503374640123, "one", 2,
         '2017-12-01 23:59:59', '2017-12-10 21:59:59', 'Open', 1503374640123),
        (1512663325123, "two", 1,
         '2017-12-02 23:59:59', '2017-12-15 19:59:59', 'Open', 1503374640123),
        (1512160044123, "three", 4,
         '2017-12-03 23:59:59', '2017-12-20 02:59:51', 'Open', 1503374640123),
        (1517480177123, "four", 6,
         '2017-12-04 23:59:59', '2017-12-22 23:59:30', 'Closed', 1503374640123),
    ]
    df = spark_session.createDataFrame(vals, columns)
    df = df.withColumn("empty", lit(None).cast(StringType()))
    df = df.withColumn("empty2", lit(None).cast(StringType()))
    df = df.sort(desc("date"))
    return df


@pytest.fixture(scope="session")
def duplicates_df(spark_session):
    """Dataframe with a duplicated row and a duplicated field"""

    columns = ["one", "two", "three"]
    vals = [
        (1503374640123, "one", 2, 3),
        (1503374640123, "one", 2, 3),
        (1503374640123, "one", 2, 4),
        (1503374644123, "one", 2, 3),
        (1503374640123, "one", 2, 3),
    ]
    df = spark_session.createDataFrame(vals, columns)
    return df


@pytest.fixture(scope="session")
def gc_instances():
    instances_names = ["cluster-datalab-01-m", "cluster-datalab-01-w-0",
                       "cluster-datalab-02-m", "cluster-datalab-02-w-0",
                       "datalake-runner", "gitlab-runner"]
    return instances_names


@pytest.fixture(scope="session")
def instances_names():
    instances_names = ["cluster-datalab-01-m", "cluster-datalab-01-w-0",
                       "cluster-datalab-01-w-1"]
    return instances_names


@pytest.fixture(scope="session")
def decimal_df(spark_session):
    """Dataframe with two decimal type columns"""
    values = [(decimal.Decimal(1.0), decimal.Decimal(1.0))]
    schema = StructType([
        StructField("one", DecimalType()),
        StructField("two", DecimalType())
    ])

    return spark_session.createDataFrame(values, schema)


@pytest.fixture(scope="session")
def timestamp_df(spark_session):
    """Dataframe with two timestamp type columns"""
    values = [(datetime.datetime(1111, 11, 11, 00, 00, 00, 000),
               datetime.datetime(1001, 10, 10, 11, 10, 11, 110))]
    schema = StructType([
        StructField("one", TimestampType()),
        StructField("two", TimestampType())
    ])
    return spark_session.createDataFrame(values, schema)


@pytest.fixture(scope="session")
def date_df(spark_session):
    """Dataframe with two date type columns"""
    values = [(datetime.date(1111, 11, 11),
               datetime.date(1001, 10, 11))]
    schema = StructType([
        StructField("one", DateType()),
        StructField("two", DateType())
    ])
    return spark_session.createDataFrame(values, schema)



from optel.datalake import descriptive_stats


def test_missing_values(me_df_timestamp):
    missing_obs = descriptive_stats.missing_values(me_df_timestamp)
    value = missing_obs.select("empty").collect()
    assert value[0][0] == 1.0


def test_describe(me_df_timestamp):
    columns = ['number']
    stats_desc = descriptive_stats.describe(me_df_timestamp, columns)
    value = stats_desc.select("number").collect()
    print(value)
    assert value[1][0] == '3.25'



import pyspark.sql.functions as sf
import logging


def missing_values(df):
    """
    Count percentage of missing values in each columns.

    .. todo:: This code is duplicated from sanity_checks

    Args:
        df (pyspark.sql.DataFrame): A spark dataframe.

    Return:
        pyspark.sql.DataFrame: A spark dataframe containing the percentage of
            missing values for each column of the input dataframe.
    """
    missing_obs = df.agg(
        *[(1 - (sf.count(c) / sf.count('*'))).
          alias(c) for c in df.columns])
    missing_obs.show()
    return missing_obs


def describe(df, columns):
    """
    Show a descriptive stats of numerical columns of a data frame.

    .. todo:: Do not write to *stdout* because it is rude.

    Args:
        df (pyspark.sql.DataFrame): A Spark Dataframe
        columns (list): A list of columns to extract descriptive stats from.

    Return:
        pyspark.sql.DataFrame: A spark Dataframe containing descriptive
            stats of the columns arg.
    """
    stats_desc = df.describe(columns)
    values = stats_desc.first()
    stats_desc_dic = values.asDict()
    logging.info("Descriptive stats of %s \n %s", df, stats_desc_dic)
    return stats_desc



from optel.datalake import sanity_checks
import pyspark.sql.functions as sf


def test_real_dups(duplicates_df):
    dedup_df = sanity_checks.real_dups(duplicates_df)
    dedup_df.show()
    nb_rows = dedup_df.count()
    nb_distinct = dedup_df.distinct().count()
    nb_dups = nb_rows - nb_distinct
    assert nb_dups == 0


def test_empty_columns(me_df_timestamp):
    df_no_empty_cols = sanity_checks.empty_columns(me_df_timestamp)
    missing_obs = df_no_empty_cols.agg(*[(1 - (sf.count(c) / sf.count('*'))).
                                         alias(c) for c in df_no_empty_cols.columns])
    values = missing_obs.first()
    dict_values = values.asDict()
    assert 1.0 not in dict_values.values()


def test_convert_decimal_to_float(decimal_df):
    df_no_decimal_cols = sanity_checks.convert_decimal_to_float(decimal_df)
    col_type = df_no_decimal_cols.dtypes
    assert 'decimal' not in col_type


def test_convert_timestamp_to_string(timestamp_df):
    df_no_timestamp_cols = sanity_checks.convert_date_to_string(timestamp_df)
    col_type = df_no_timestamp_cols.dtypes
    assert 'timestamp' not in col_type


def test_convert_date_to_string(date_df):
    df_no_date_cols = sanity_checks.convert_date_to_string(date_df)
    col_type = df_no_date_cols.dtypes
    assert 'date' not in col_type



"""Package main definition."""


from pkg_resources import get_distribution, DistributionNotFound


__project__ = 'optel.datalake'
__version__ = None  # required for initial installation

try:
    distribution = get_distribution(__project__)
    __version__ = distribution.version

except DistributionNotFound:
    # This will happen if the package is not installed.
    # For more informations about development installation, read about
    # the 'develop' setup.py command or the '--editable' pip option.
    # Note that development installations may break other packages from
    # the same implicit namespace
    # (see https://github.com/pypa/packaging-problems/issues/12)
    __version__ = '(local)'
else:
    pass
    # make shortcut import here



from pyspark.sql import SparkSession


def get_spark(name):
    """
    Get a Spark session by name.

    The Spark session is reused if it already exists.

    Args:
        name (str): The name of the Spark session.

    Returns:
        pyspark.sql.SparkSession: Ready to use session.
    """
    spark = SparkSession \
        .builder \
        .appName(name) \
        .getOrCreate()
    return spark


def default_spark():
    """
    Provide the default Spark session.

    Returns:
        pyspark.sql.SparkSession: A Spark session named *Optel Data Lake*.
    """
    return get_spark('Optel Data Lake')



from pyspark.sql import functions as sf
from pyspark.sql.types import DecimalType
from tzlocal import get_localzone


def convert_ms_timestamp(table, date_columns):
    """
    Convert unix timestamp (ms) to date.

    Args:
        table (str): Spark dataframe
        date_columns (list): A list of columns to convert
    """
    localtz = get_localzone()
    for date_column in date_columns:
        df_tmp = table.withColumn(
            date_column, (sf.col(date_column)/1000).cast(DecimalType(12, 0)))
        df_tmp = df_tmp.withColumn(
            date_column, df_tmp[date_column].cast('timestamp'))
        df_tmp = df_tmp.withColumn(date_column, sf.to_utc_timestamp(
            df_tmp[date_column], str(localtz)))
        table = df_tmp
    return df_tmp


def find_key(dic, val):
    """
    Return the key of dictionary dic given the value.

    Args:
        dic (dict): A dictionary.
        val (str): A string to search for in the dictionary.

    Returns:
        list: A list of keys (columns) containing a
            specified value (val).
    """
    empty_cols = []
    for key, value in dic.items():
        if value == val:
            empty_cols.append(key)
        else:
            pass
    return empty_cols



import click
import logging
from optel.datalake import cluster_events
from optel.datalake import __version__, __project__


@click.group()
@click.version_option(__version__)
def main():
    logging.basicConfig(filename='info.log',
                        level=logging.INFO,
                        filemode='w')


@main.group()
def cluster():
    pass


@cluster.command()
@click.option('--project', help='Project where the instances are')
@click.option('--zone', help='Zone where the project exist')
@click.option('--cluster', help='Cluster to boot')
def boot(project, zone, cluster):
    instances = cluster_events.get_instances_names(project, zone, cluster)[0]
    cluster_events.boot_cluster(instances, project, zone)


@cluster.command()
@click.option('--project', help='Project where the instances are')
@click.option('--zone', help='Zone where the project exist')
@click.option('--cluster', help='Cluster to kill')
def kill(project, zone, cluster):
    instances = cluster_events.get_instances_names(project, zone, cluster)[0]
    cluster_events.shutdown_cluster(instances, project, zone, cluster)


if __name__ == '__main__':
    main(progname=__project__)



import sys


if __name__ == '__main__' and not __package__:
    # This should never happen when installed from pip (setup.py).
    # This workaround is NOT bulletproof, rather brittle as many edge
    # cases are not covered
    # See http://stackoverflow.com/a/28154841/2479038
    print('warning: running package directly, risking ImportError',
          file=sys.stderr)


from optel.datalake import __project__  # noqa
from optel.datalake.cli import main  # noqa


if __name__ == '__main__':
    sys.exit(main(prog_name=__project__))



import googleapiclient.discovery
import subprocess


def boot_cluster(instances_names, project, zone):
    """
    Boot a dataproc cluster on GC.

    Args:
        instances_names (list): A list of instances to boot.
        project (str): The project where to boot instances.
        zone (str): The zone of the project.

    Example:
        compute.instances().start(project=project, instance=instance[0],
        zone=zone.)
    """
    compute = googleapiclient.discovery.build('compute', 'v1')

    print("Booting", instances_names)
    for instance in instances_names:
        compute.instances().start(
            project=project, zone=zone, instance=instance).execute()


def shutdown_cluster(instances_names, project, zone, cluster):
    """
    Shutdown a dataproc cluster if no jobs are running.

    Args:
        instances_names (list): A list instances to shutdown.
        project (str): The project where to boot instances.
        zone (str): The zone of the project.
        cluster (str): The cluster to shutdown
    """
    compute = googleapiclient.discovery.build('compute', 'v1')

    print("Checking if a job is running")
    status = subprocess.check_output(["gcloud", "dataproc", "jobs", "list",
                                      "--cluster", cluster,
                                      "--region", zone[:-2],
                                      "--state-filter", "active"])
    print(status)
    if b"JOB_ID" in status:
        print("Job running, not shutting down the cluster")

    else:
        print("No jobs running, shutting down cluster")
        for instance in instances_names:
            compute.instances().stop(
                project=project, zone=zone, instance=instance).execute()


def get_instances_names(project, zone, cluster):
    """
    Get instance names based on cluster name and their status.

    Args:
        project (str): The project where to get the names.
        zone (str): The zone of the project.
        cluster (str): The cluster to get the instances names.

    Returns:
        list: Selected instances based on cluster.

    """
    compute = googleapiclient.discovery.build('compute', 'v1')

    project_objects = compute.instances().list(
        project=project, zone=zone).execute()

    cluster_status = compute.instances().get(
        project=project, zone=zone,
        instance=project_objects['items'][0]['name']).execute()['status']
    instances_names = []
    ninstances = len(project_objects['items'])
    for instance in range(ninstances):
        instances_names.append(project_objects['items'][instance]['name'])
    selected_instances = []
    for instance in instances_names:
        if cluster in instance:
            selected_instances.append(instance)
    return selected_instances, cluster_status



from optel.datalake import ingest

import pytest


@pytest.mark.skip(reason="Not working on runner, cant set SPARK_HOME")
def test_write_to_elastic(me_df_timestamp):
    destination = 'test-datalake'
    username = 'elastic-bot-write'
    password = 'M2cscea.Istt'
    nodes = 'https://6f3b94ab79c327269a56379a235bdf86.us-central1.gcp.cloud.es.io:9243'
    ingest.write_to_elastic(
        me_df_timestamp, destination, nodes, username, password)



from optel.datalake import transform
from optel.datalake import descriptive_stats


def test_convert_ms_timestamp(me_df_timestamp):
    df_transformed = transform.convert_ms_timestamp(
        me_df_timestamp, ["date", "date2"])
    test_date = df_transformed.select("date").collect()[0][0]
    assert(str(test_date) == "2018-02-01 10:16:17")


def test_missing_values(me_df_timestamp):
    df = descriptive_stats.missing_values(me_df_timestamp)
    null_value = df.select("empty").collect()[0][0]
    assert(null_value == 1.0)



from optel.datalake.config.sparksession import default_spark
from optel.datalake import sanity_checks


def load_sql_tables(database, tables, server, username, password, spark=None):
    """
    Load one table or a table schema from SQL server.

    Args:
        database (str): Database from which to fetch tables.
        tables (list): A list of tables to fetch.
        server (str): The server from which to fetch data.
        username (str): The username used to connect to the server.
        password (str): The password of the aforementioned username.
        spark (pyspark.sql.SparkSession): Optional Spark session to use.  This
            will default to the session provided by
            :func:`optel.datalake.config.sparksession.default_spark`

    Returns:
        dict: Data frames keyed by table name
    """
    spark = spark or default_spark()
    dic_of_df = dict()
    for table in tables:
        spark_df = spark.read \
            .format("jdbc") \
            .option("url", server) \
            .option("databaseName", database) \
            .option("dbtable", table) \
            .option("user", username) \
            .option("password", password) \
            .option("driver", "com.microsoft.sqlserver.jdbc.SQLServerDriver") \
            .option("header", "true") \
            .option("nullValue", "0") \
            .option("encrypt", "false") \
            .load()
        dic_of_df[table] = spark_df

    return dic_of_df


def write_to_elastic(table, destination, nodes, username, password):
    """
    Write table to Elastic search.

    Args:
        table (pyspark.sql.DataFrame): The name of the spark dataframe that
            will be written to Elastic.
        destination (str): The name of the index under which the dataframe
            will be written.
        nodes (str): Elastic search nodes where to write table.
        username (str): Elasticsearch user with write access.
        password (str): Elasticsearch user password.
    """
    table = sanity_checks.convert_decimal_to_float(table)
    table = sanity_checks.convert_date_to_string(table)
    resource_name = '{}/table'.format(destination)
    table.write \
        .format("org.elasticsearch.spark.sql") \
        .option("es.mapping.date.rich", "true") \
        .mode("overwrite") \
        .option("es.index.auto.create", "true") \
        .option("es.resource", resource_name) \
        .option("es.nodes", nodes) \
        .option("es.net.ssl.protocol", "true") \
        .option("es.nodes.wan.only", "true") \
        .option("es.net.http.auth.user", username) \
        .option("es.field.read.empty.as.null", "yes") \
        .option("es.net.http.auth.pass", password) \
        .save()


def read_from_elastic(index, nodes, username, password, spark=None):
    """
    Read a table from Elasic Search.

    Args:
        index (str): The name of the elastic search index to read.
        spark (pyspark.sql.SparkSession): Optional Spark session to use.  This
            will default to the session provided by
            :func:`optel.datalake.config.sparksession.default_spark`
    """
    spark = spark or default_spark()
    loaded_es_index = spark.read \
        .format("org.elasticsearch.spark.sql") \
        .option("es.nodes", nodes) \
        .option("es.net.ssl.protocol", "true") \
        .option("es.nodes.wan.only", "true") \
        .option("es.net.http.auth.user", username) \
        .option("es.net.http.auth.pass", password) \
        .load(index + "/table")
    return loaded_es_index


def write_csv(df, destination):
    """
    Write a csv table to a destination.

    Args:
        df (pyspark.sql.DataFrame): TODO
        destination (str): CSV file path.
    """
    df.write \
        .format("csv") \
        .mode("overwrite") \
        .option("header", "true") \
        .option("inferSchema", "true")\
        .csv(destination)


def write_parquet(df, destination):
    """
    Write a Spark dataframe to a parket file.

    Args:
        df (pyspark.sql.DataFrame): The name of the Spark dataframe.
        destination (str): The destination file path where to write the
            dataframe *parquet* file.
    """
    df.write \
        .mode("overwrite") \
        .parquet(destination)


def read_csv(source, spark=None):
    spark = spark or default_spark()
    df = spark.read \
        .format("csv") \
        .option("header", "true") \
        .option("inferSchema", "true") \
        .load(source)
    return df


def read_parquet(source, spark=None):
    """
    Read parquet files into a spark dataframe.

    Args:
        source (str): Source *parquet* file.
        spark (pyspark.sql.SparkSession): Optional Spark session to use.  This
            will default to the session provided by
            :func:`optel.datalake.config.sparksession.default_spark`

    Returns:
        pyspark.sql.DataFrame: Properly initialized data frame.
    """
    spark = spark or default_spark()
    result_df = spark.read.parquet(source)
    return result_df


def get_df_from_source(base_source, tables):
    """
    Make data frames from a folder of *parquet* files.

    Args:
        base_source (str): Path to the folder containing *parquet* files.
        tables: List of parquet file names.

    Returns:
        dict: Mapping of :class:`pyspark.sql.DataFrame` objects indexed by
            file name.
    """
    dic_of_df = {}
    for table in tables:
        source = (base_source + table)
        dic_of_df[table] = read_parquet(source)
    dic_of_df[table].show(n=2)
    return dic_of_df



#!/usr/bin/python3
# coding: utf8


"""This scripts lays building blocks for development tooling."""


from distutils.version import LooseVersion
from pip.req import parse_requirements
from pip.download import PipSession
import email.utils
import errno
import fnmatch
import os
import platform
import setuptools
import shutil
from urllib.parse import urlparse

__all__ = (
    'ProjectMetadata',
    'Clean',

    'Documentation',
)


PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))

PYTHON_VERSION = platform.python_version_tuple()

DEFAULT_VERSION = '0.1'


class ProjectMetadata(object):
    """
    This hold distribution information primarily used at build time.

    Properties of this class are intended to roughly fit the
    `metadata
    <https://docs.python.org/2/distutils/setupscript.html#additional-meta-data>`_
    expected by keyword arguments by the :func:`setuptools.setup` function.

    Since we use *setuptools*, you may wish to read about
    `new and changed setup keywords
    <http://setuptools.readthedocs.io/en/latest/setuptools.html?highlight=keywords#new-and-changed-setup-keywords>`.
    """

    def __init__(self):
        """Initialize some cached or internal storage of data."""
        self._version = None

    @property
    def name(self):
        """
        Name of the package.

        Note:
            This field is required by :func:`setuptools.setup`

        Returns:
             str: A short string
        """
        value = 'optel.datalake'
        return self._ensure_short_string(value)

    @property
    def description(self):
        """
        Short, summary description of the package.

        Returns:
             str: A short string
        """
        value = 'Common tools used in the data refinery'
        return self._ensure_short_string(value)

    @property
    def long_description(self):
        """
        Returns:
             str: A string
        """
        return open('README.rst').read()

    @property
    def version(self):
        """
        Version of this release.

        Note:
            This field is required by :func:`setuptools.setup`

        Returns:
             str: A short string
        """
        if self._version is None:
            try:
                self._version = str(LooseVersion(read_file('VERSION'))).strip()
            except (OSError, IOError):
                self._version = str(LooseVersion(os.getenv('_VERSION',
                                                           DEFAULT_VERSION)))

        return self._ensure_short_string(self._version)

    @property
    def url(self):
        """
        Home page for the package.

        Note:
            This field is required by :func:`setuptools.setup`

        Returns:
             str: A URL
        """
        value = 'https://www.optelgroup.com/'
        return self._ensure_url(value)

    @property
    def download_url(self):
        """
        Location where the package may be downloaded.

        Returns:
             str: A URL
        """
        value = '{!s}/repository/master/archive.tar.gz'.format(self.url)
        return self._ensure_url(value)

    @property
    def license(self):
        """
        License for the package.

        Note:
            The "license" field is a text indicating the license
            covering the package where the license is not a selection
            from the "License" Trove classifiers. See the :attr:`~classifiers`
            field. Notice that there's a "licence" distribution option
            which is deprecated but still acts as an alias for license.

        Returns:
             str: A short string
        """
        value = 'LGPL'
        return self._ensure_short_string(value)

    @property
    def keywords(self):
        """
        A whitespace-separated list of keywords about this project.

        I have been able to find documentation about this parameter but it
        is clearly used in the setuptools documentation in the example
        provided for `basic usage
        <http://setuptools.readthedocs.io/en/latest/setuptools.html?highlight=keywords#basic-use>`
        """
        return ''

    @property
    def author(self):
        """
        Package author’s name.

        Either the author or the maintainer must be identified. If
        maintainer is provided, distutils lists it as the author in
        PKG-INFO.

        Returns:
             str: A short string
        """
        value = 'Optel Business Technologies Team'
        return self._ensure_short_string(value)

    @property
    def author_email(self):
        """
        Email address of the package author.

        See also :attr:`~author`

        Returns:
             str: An email address
        """
        value = 'btechsupport@optelgroup.com'
        return self._ensure_email_address(value)

    @property
    def classifiers(self):
        """
        A list of classifiers.

        Feel free to consult
        `all possibles classifiers
        <https://pypi.python.org/pypi?%3Aaction=list_classifiers>`.

        Returns:
             List[str]: A list of short strings
        """
        value = ['Development Status :: 4 - Beta',
                 'Intended Audience :: Developers',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Utilities']
        return [self._ensure_short_string(string) for string in value]

    @property
    def packages(self):
        """
        A list of packages to distribute (and all their submodules).

        It's usage is explained in
        `distutils documentation
        <https://docs.python.org/2/distutils/setupscript.html#listing-whole-packages>`

        You may also read about :func:`setuptools.find_packages` function and
        `it's usage
        <http://setuptools.readthedocs.io/en/latest/setuptools.html?highlight=keywords#using-find-packages>`.

        Returns:
             List[str]: A list of strings
        """
        find_packages = setuptools.PEP420PackageFinder.find
        value = find_packages(str(PROJECT_ROOT), exclude=['doc',
                                                          'test*',
                                                          'requirements'])
        return [self._ensure_short_string(package) for package in value]

    @property
    def install_requires(self):
        """
        Production dependancies.

        A string or list of strings specifying what other distributions need
        to be installed when this one is.

        It's usage is explained in `setuptools documentation
        <http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-dependencies>`

        Returns:
             List[str]: A list of short strings
        """
        install_reqs = parse_requirements('requirements.txt',
                                          session=PipSession())
        reqs = [str(ir.req) for ir in install_reqs]
        return list(reqs)

    @property
    def tests_require(self):
        """
        Automated tests dependancies.

        If your project’s tests need one or more additional packages besides
        those needed to install it, you can use this option to specify them.

        It's usage is explained in `setuptools documentation
        <http://setuptools.readthedocs.io/en/latest/setuptools.html#new-and-changed-setup-keywords>`

        Returns:
             List[str]: A list of short strings
        """
        return []

    @property
    def setup_requires(self):
        """
        A string or list of strings specifying what other distributions need to
        be present in order for the setup script to run. setuptools will
        attempt to obtain these (even going so far as to download them using
        EasyInstall) before processing the rest of the setup script or
        commands. This argument is needed if you are using distutils extensions
        as part of your build process; for example, extensions that process
        `setup()` arguments and turn them into EGG-INFO metadata files.

        Note:
            Projects listed in setup_requires will NOT be automatically
            installed on the system where the setup script is being run. They
            are simply downloaded to the ./.eggs directory if they’re not
            locally available already. If you want them to be installed, as
            well as being available when the setup script is run, you should
            add them to install_requires and setup_requires.

        Returns:
             List[str]: A list of short strings
        """
        return []

    @property
    def extras_require(self):
        """
        A dictionary mapping names of `extras` (optional features of your
        project) to strings or lists of strings specifying what other
        distributions must be installed to support those features.

        See Also:
            http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-extras-optional-features-with-their-own-dependencies

        Returns:
            Dict: Lists of dependencies indexed by the feature they enable.
        """
        return {}

    def setup(self):
        """Run :func:`setuptools.setup` using :func:~`raw`."""
        return setuptools.setup(**self.raw())

    def raw(self):
        """
        Create a generic dict representation.

        Provides a uniform access to all distribution data as python
        primitives, exactly the same one would provide to
        :func:`setuptools.setup`

        Returns:
            dict: Keyword arguments for :func:`setuptools.setup`.
        """
        return dict(name=self.name,
                    version=self.version,
                    description=self.description,
                    long_description=self.long_description,
                    url=self.url,
                    license=self.license,
                    author=self.author,
                    author_email=self.author_email,
                    classifiers=self.classifiers,
                    packages=self.packages,
                    include_package_data=True,
                    entry_points={
                        'console_scripts': [
                            '{module!s}={module!s}.cli:main'.format(
                                module=self.name)
                        ]
                    },
                    cmdclass={'doc': Documentation,
                              'clean': Clean},
                    install_requires=self.install_requires,
                    tests_require=self.tests_require,
                    setup_requires=self.setup_requires,
                    extras_require=self.extras_require)

    def __str__(self):
        """Provide a human-readable representation."""
        import pprint
        return pprint.pformat(self.raw(),)

    @staticmethod
    def _ensure_short_string(string):
        string = str(string)
        assert len(string) <= 200
        return string

    @staticmethod
    def _ensure_long_string(string):
        string = str(string)
        return string

    @staticmethod
    def _ensure_email_address(string):
        string = str(string)
        _, email_str = email.utils.parseaddr(str(string))
        assert email_str == str(string)
        return string

    @staticmethod
    def _ensure_url(string, qualifying=None):
        string = str(string)
        min_attributes = ('scheme', 'netloc')

        qualifying = min_attributes if qualifying is None else qualifying
        token = urlparse(str(string))
        valid = all([getattr(token, qualifying_attr)
                     for qualifying_attr in qualifying])
        assert valid
        return string


class Clean(setuptools.Command):
    """Custom clean command to tidy up the project."""

    description = 'Custom clean command to tidy up the project.'
    user_options = []
    default_patterns = ['build',  'dist', '*.egg-info', '*.egg', '*.pyc',
                        '*.pyo', '*~', '__pycache__', '.tox', '.coverage',
                        'htmlcov']

    def initialize_options(self):
        """Set default values for options."""
        pass

    def finalize_options(self):
        """Post-process options."""
        pass

    def run(self):
        """Run command."""
        for pattern in self.default_patterns:
            self._remove_matching_files(pattern)

    def _remove_matching_files(self, pattern):
        for path in find_files(pattern):
            self.announce('Cleaning path {!s}'.format(path), 2)
            shutil.rmtree(path, ignore_errors=True)


class Documentation(setuptools.Command):
    """
    Make the documentation (without the *Make* program).

    .. note::

        This command will not allow any warning from `Sphinx`_, treating
        them as errors.

    .. _Sphinx: http://www.sphinx-doc.org/
    """

    description = 'create documentation distribution'
    user_options = [
        ('builder=', None, 'documentation output format '
                           '[default: html]'),
        ('dist-dir=', None, 'directory to put final built distributions in '
                            '[default: dist/doc]'),
        ('build-dir=', None, 'temporary directory for creating the '
                             'distribution '
                             '[default: build/doc]'),
        ('src-dir=', None, 'documentation source directory '
                           '[default: doc]'),
        ('ignore-warnings', None, 'let the documentation be built even if '
                                  'there are warnings [default: False]')
    ]

    targets = {
        'html': {
            'comment': 'The HTML pages are in {build_dir!s}.',
        },
        'dirhtml': {
            'comment': 'The HTML pages are in {build_dir!s}.',
        },
        'singlehtml': {
            'comment': 'The HTML pages are in {build_dir!s}.',
        },
        'latex': {
            'comment': 'The LaTeX files are in {build_dir!s}.',
        },
    }

    def initialize_options(self):
        """Set default values for options."""
        import os
        # Each user option must be listed here with their default value.
        # project_directory = PROJECT_ROOT

        self.builder = 'html'
        self.dist_dir = os.path.join(PROJECT_ROOT, 'dist', 'doc')
        self.build_dir = os.path.join(PROJECT_ROOT, 'build', 'doc')
        self.src_dir = os.path.join(PROJECT_ROOT, 'doc')
        self.ignore_warnings = False

    def finalize_options(self):
        """Post-process options."""
        self.ignore_warnings = bool(self.ignore_warnings)
        if self.builder in self.targets:
            self._actual_targets = [self.builder]
        elif self.builder == 'all':
            self._actual_targets = self.targets.keys()
        else:
            self._actual_targets = []

        # package_dir may be None, in that case use the current directory.
        metadata = self.distribution.metadata
        self.code_dir = os.path.abspath((self.distribution.package_dir
                                         or {'': metadata.name})[''])
        self._canonical_directories()

        self.announce(
            'Building {!s} documentation'.format(self.builder), 2)
        self.announce(
            '  using source at "{!s}"'.format(self.src_dir), 2)
        self.announce(
            '  putting work files at "{!s}"'.format(self.build_dir), 2)
        self.announce(
            '  distributing documentation at "{!s}"'.format(self.dist_root), 2)
        self.announce(
            '  looking for code at "{!s}"'.format(self.code_dir), 2)

    def run(self):
        """Run command."""
        result_dirs = {}
        for target in self._actual_targets:
            result_dirs[target] = self._build_doc(target)
        for target, result_dir in result_dirs.items():
            self.announce('Documentation target "{!s}" : {!s}'.format(
                target,
                self.targets[target]['comment'].format(
                    build_dir=result_dir)), 2)

    def _canonical_directories(self):
        self.dist_root = os.path.abspath(mkdir_p(self.dist_dir))
        self.build_dir = os.path.abspath(mkdir_p(self.build_dir))
        self.src_dir = os.path.abspath(mkdir_p(self.src_dir))

    def _build_source(self):
        import sphinx.apidoc
        build_dir = self.src_dir

        # metadata contains information supplied in setup()
        metadata = self.distribution.metadata

        sphinx_apidoc_opts = [
            '',
            '--module-first',   # Put module documentation before submodule
            # documentation
            '--doc-project', metadata.name,
            '--output-dir', build_dir,
            '--maxdepth', 4,
            '--module-first',
            '--implicit-namespaces',
            PROJECT_ROOT,  # All the project is sourced
            'setup.py',
        ]

        excluded_subdirectories = [directory
                                   for directory in os.listdir(PROJECT_ROOT)
                                   if os.path.isdir(directory)
                                   and directory not in [metadata.name,
                                                         'optel',
                                                         'test']]

        sphinx_apidoc_opts.extend(excluded_subdirectories)

        canonical_opts = [str(arg) for arg in sphinx_apidoc_opts]

        self.announce('  Invoking sphinx : '
                      '"{!s}"'.format(' '.join(canonical_opts)), 2)
        sphinx.apidoc.main([str(arg) for arg in sphinx_apidoc_opts])

        return build_dir

    def _build_doc(self, target):
        import shutil
        import sphinx

        rst_src = self._build_source()
        metadata = self.distribution.metadata

        build_dir = os.path.join(self.build_dir, target)
        dist_dir = os.path.join(self.dist_root, target)

        cached_directory = os.path.join(os.path.dirname(build_dir), 'doctrees')

        all_sphinx_opts = [
            '',
            '-b', target,             # builder to use; default is html
            '-d', cached_directory,   # path for the cached doctree files
            '-n',                     # warn about all missing references
            '-q',                     # no output on stdout, warning on stderr
            '-T',                     # show full traceback on exception
            '-D', 'project={!s}'.format(metadata.name),
            '-D', 'version={!s}'.format(metadata.version)
        ]

        if not self.ignore_warnings:
            all_sphinx_opts.append('-W')

        all_sphinx_opts.append(rst_src)
        all_sphinx_opts.append(build_dir)

        sphinx.build_main([str(arg) for arg in all_sphinx_opts])
        shutil.rmtree(str(dist_dir), ignore_errors=True)
        mkdir_p(os.path.dirname(dist_dir))
        shutil.copytree(str(build_dir), str(dist_dir))

        return dist_dir


def read_file(file_path):
    """Read text from file relative to the project root."""
    return open(os.path.join(PROJECT_ROOT, str(file_path))).read()


def list_from_file(file_path):
    content = read_file(file_path)
    return [line.strip() for line in content.splitlines() if line]


def find_files(pattern, directory=PROJECT_ROOT):
    """Generate file names matching pattern recursively found in directory."""
    for root, dirs, files in os.walk(str(directory)):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename


def mkdir_p(path):
    """
    Behave like ``mkdir -b``

    Args:
        path (str):

    Returns:
        str: resulting path
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise
    return path


def main():
    """Main entry point."""
    meta = ProjectMetadata()
    meta.setup()


if __name__ == '__main__':
    main()
